/* Cree le : 16 mars 2007
 */

/*
 * dds.c -- dds texture loader
 * last modification: 16/03/07 (fr) by Antony Martin
 *
 * Copyright (c) 2005-2006 David HENRY
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * gcc -Wall -ansi -L/usr/X11R6/lib -lGL -lGLU -lglut dds.c -o dds
 */



#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEString.h>

#include <SCE/interface/SCEDDSLoader.h>


#define MAKEFOURCC(ch0, ch1, ch2, ch3) \
  (GLuint)( \
    (((GLuint)(GLubyte)(ch3) << 24) & 0xFF000000) | \
    (((GLuint)(GLubyte)(ch2) << 16) & 0x00FF0000) | \
    (((GLuint)(GLubyte)(ch1) <<  8) & 0x0000FF00) | \
     ((GLuint)(GLubyte)(ch0)        & 0x000000FF) )

#define FOURCC_DXT1 MAKEFOURCC('D', 'X', 'T', '1')
#define FOURCC_DXT3 MAKEFOURCC('D', 'X', 'T', '3')
#define FOURCC_DXT5 MAKEFOURCC('D', 'X', 'T', '5')



void* SCE_DDS_LoadFromFile(FILE *fp, const char *filename)
{
    SCE_SDDSurfaceDesc ddsd;
    SCE_SDDSImage *texinfo = NULL;
    char magic[4];
    long bufferSize, curr, end;
    
    
    /* read magic number and check if valid .dds file */
    fread (&magic, sizeof (char), 4, fp);
    
    if (strncmp (magic, "DDS ", 4) != 0)
    {
        Logger_Log(SCE_BAD_FORMAT);
        Logger_LogMsg("the file '%s' doesn't appear to be "
                      "a valid .dds file!\n", filename);
        
        return (NULL);
    }
    
    /* get the surface descriptor */
    fread (&ddsd, sizeof (ddsd), 1, fp);
    
    texinfo = SCE_malloc(sizeof *texinfo);
    memset(texinfo, 0, sizeof *texinfo);
    texinfo->width = ddsd.width;
    texinfo->height = ddsd.height;
    texinfo->n_levels = ddsd.n_levels;
    
    switch (ddsd.format.fourCC)
    {
        case FOURCC_DXT1:
            /* DXT1's compression ratio is 8:1 */
            texinfo->format = GL_COMPRESSED_RGBA_S3TC_DXT1_EXT;
            break;
    
        case FOURCC_DXT3:
            /* DXT3's compression ratio is 4:1 */
            texinfo->format = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
            break;
    
        case FOURCC_DXT5:
            /* DXT5's compression ratio is 4:1 */
            texinfo->format = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
            break;
    
        default:
            /* bad fourCC, unsupported or bad format */
            Logger_Log(SCE_BAD_FORMAT);
            Logger_LogMsg("the file '%s' doesn't appear to be "
                          "compressed using DXT1, DXT3, or DXT5! [%i]",
                          filename, ddsd.format.fourCC);
            SCE_free(texinfo);
            
            return (NULL);
    }
    
    /* calculate pixel data size */
    curr = ftell (fp);
    fseek (fp, 0, SEEK_END);
    end = ftell (fp);
    fseek (fp, curr, SEEK_SET);
    bufferSize = end - curr;
    
    /* read pixel data with mipmaps */
    texinfo->texels = SCE_malloc(bufferSize * sizeof *texinfo->texels);
    fread(texinfo->texels, sizeof *texinfo->texels, bufferSize, fp);
    
    
    return (texinfo);
}
