/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 19/05/2007
   derniere modification : 19/06/2007 */


/**
IMPORTANT :
a reviser : la gestion des textures, voir si tout est bien implemente
*/

#include <SCE/SCEMinimal.h>

#include <SCE/core/SCECTexture.h>
#include <SCE/core/SCECFramebuffer.h>
#include <SCE/core/SCECShader.h>
#include <SCE/core/SCECMaterial.h>

#include <SCE/interface/SCEResources.h>
#include <SCE/interface/SCETexture.h>
#include <SCE/interface/SCEShaders.h>

#include <SCE/interface/SCEMaterial.h>



/* revise le 18/06/2007 */
SCE_SMaterial* SCE_Material_Create(void)
{
    SCE_SMaterial *mat = NULL;
    int i, j;
    
    mat = SCE_malloc(sizeof *mat);
    if(mat == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    mat->cmat = SCE_CCreateMaterial();
    if(mat->cmat == NULL)
    {
        SCE_free(mat);
        Logger_LogSrc();
        return (NULL);
    }
    
    for(i=0; i<SCE_NUM_MAPS; i++)
    {
        mat->maps[i] = SCE_malloc(SCE_CGetInteger(GL_MAX_TEXTURE_UNITS) *
                                  sizeof *mat->maps[i]);
        if(mat->maps[i] == NULL)
        {
            Logger_LogSrc();
            
            for(i--; i>-1; i--)
                SCE_free(mat->maps[i]);
            
            SCE_CDeleteMaterial(mat->cmat);
            SCE_free(mat);
            
            return (NULL);
        }
        
        for(j=0; j<SCE_CGetInteger(GL_MAX_TEXTURE_UNITS); j++)
        {
            mat->maps[i][j].tex = NULL;
            mat->maps[i][j].actived = SCE_FALSE;
            mat->maps[i][j].user = SCE_FALSE;
        }
    }
    
    mat->shader = NULL;
    mat->shader_actived = SCE_FALSE;
    mat->shader_user = SCE_FALSE;
    
    return (mat);
}

/* revise le 18/06/2007 */
void SCE_Material_Delete(SCE_SMaterial *mat)
{
    int i, j;
    
    if(mat != NULL)
    {
        /* pour chaque type de texture */
        for(i=0; i<SCE_NUM_MAPS; i++)
        {
            /* pour chaque unite de texture */
            for(j=0; j<SCE_CGetInteger(GL_MAX_TEXTURE_UNITS); j++)
            {
                /* on ne la detruit que si elle n'est pas a l'utilisateur */
                if(!mat->maps[i][j].user)
                    SCE_Texture_Delete(mat->maps[i][j].tex);
            }
            
            SCE_free(mat->maps[i]);
        }
        
        if(!mat->shader_user)
            SCE_Shader_Delete(mat->shader);
        
        SCE_CDeleteMaterial(mat->cmat);
        
        SCE_free(mat);
    }
}



void SCE_Material_SetColor(SCE_SMaterial *mat, SCEenum type,
                           float r, float g, float b, float a)
{
    if(mat != NULL)
        SCE_CSetColorMaterial(mat->cmat, type, r, g, b, a);
}

void SCE_Material_SetColorv(SCE_SMaterial *mat, SCEenum type, float *color)
{
    if(mat != NULL)
        SCE_CSetColorMaterialv(mat->cmat, type, color);
}


/* ajoute le 18/06/2007 */
int SCE_Material_GenTextureUnit(SCE_SMaterial *mat, int type)
{
    int i;
    
    for(i=0; i<SCE_CGetInteger(GL_MAX_TEXTURE_UNITS); i++)
    {
        if(mat->maps[type][i].tex == NULL)
            return (i);
    }
    
    return (-1);
}


/* revise le 18/06/2007 */
SCEuint SCE_Material_AddTexture_(SCE_SMaterial *mat, SCEenum type,
                                 SCEuint unit, SCE_STexture *tex, int user)
{
    #ifdef __SCE_DEBUG__
    if(mat == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    /** TODO: a deplacer dans une fonction du coeur (a part) (voir aussi Mesh) */
    if(unit >= (unsigned int)SCE_CGetInteger(GL_MAX_TEXTURE_UNITS))
    {
        Logger_Log(SCE_INVALID_SIZE);
        Logger_LogMsg("the %dth texture unit is not supported by your hardware",
                      unit);
        return (SCE_ERROR);
    }
    
    /* si une texture est deja presente ici, et qu'elle n'appartient pas
       a l'utilisateur */
    if(mat->maps[type][unit].tex != NULL && !mat->maps[type][unit].user)
    {
        /* on la supprime :D bourrin ... alors on laisse un message */
        Logger_PrintMsg("erasing of the texture named '%s' in the material.\n",
                      SCE_Resource_GetName(mat->maps[type][unit].tex->data[0]));
        SCE_Texture_Delete(mat->maps[type][unit].tex);
    }
    
    mat->maps[type][unit].tex = tex;
    mat->maps[type][unit].actived = SCE_TRUE;
    mat->maps[type][unit].user = user;
    
    /* on renvoie l'unite de tex sous laquelle a ete enregistree la texture */
    return (unit);
}

/* revise le 18/06/2007 */
void SCE_Material_SetShader_(SCE_SMaterial *mat, SCE_SShader *shader, int user)
{
    if(mat != NULL)
    {
        /* NOTE p-e a deplacer dans LoadShader pour moins fragmenter la mem ? */
        /* si le materiau a deja un shader, on le supprime */
        if(mat->shader != NULL && !mat->shader_user)
            SCE_Shader_Delete(mat->shader);
        
        /* assignation du nouveau shader */
        mat->shader = shader;
        mat->shader_actived = SCE_TRUE;
        mat->shader_user = user;
    }
}



/* revise le 18/06/2007 */
SCEuint SCE_Material_LoadTexture(SCE_SMaterial *mat, SCEenum type, SCEuint unit,
                                 const char *name, int use_mipmap)
{
    SCE_STexture *tex = NULL;
    
    /* creation & chargement de la texture */
    tex = SCE_Texture_CreateFromFile(name, use_mipmap);
    if(tex == NULL)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    /* ajout de la texture au materiau */
    return (SCE_Material_AddTexture_(mat, type, unit, tex, SCE_FALSE));
}

/* revise le 18/06/2007 */
SCEuint SCE_Material_LoadCubemap(SCE_SMaterial *mat, SCEuint unit,
                                 const char *px, const char *nx, const char *py,
                                 const char *ny, const char *pz, const char *nz,
                                 int use_mipmap)
{
    SCEuint result;
    SCE_STexture *tex = NULL;
    
    tex = SCE_Texture_Create(SCE_TEX_CUBE, 0, 0);
    if(tex == NULL)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    SCE_Texture_AddImage(tex, SCE_TEX_POSX, px);
    SCE_Texture_AddImage(tex, SCE_TEX_NEGX, nx);
    SCE_Texture_AddImage(tex, SCE_TEX_POSY, py);
    SCE_Texture_AddImage(tex, SCE_TEX_NEGY, ny);
    SCE_Texture_AddImage(tex, SCE_TEX_POSZ, pz);
    SCE_Texture_AddImage(tex, SCE_TEX_NEGZ, nz);
    
    if(SCE_Texture_Build(tex, use_mipmap) < 0)
    {
        SCE_Texture_Delete(tex);
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    if(SCE_Texture_Update(tex) < 0)
    {
        SCE_Texture_Delete(tex);
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    /* ajout de la texture au materiau */
    result = SCE_Material_AddTexture_(mat, SCE_CUBE_MAP, unit, tex, SCE_FALSE);
    if(result == (unsigned int)SCE_ERROR) /* NOTE: un peu caca non ? */
        Logger_LogSrc();
    
    return (result);
}


int SCE_Material_LoadShader(SCE_SMaterial *mat, const char *vs, const char *ps)
{
    SCE_SShader *shd = NULL;
    
    shd = SCE_Shader_CreateFromFile(vs, ps);
    if(shd == NULL)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    if(SCE_Shader_Build(shd) < 0)
    {
        SCE_Shader_Delete(shd);
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    SCE_Material_SetShader(mat, shd);
    
    return (SCE_OK);
}


/* ajoute le 14/06/2007 */
/* revise le 19/06/2007 */
/** fonction a tester */
void SCE_Material_DeleteTexture(SCE_SMaterial *mat, int type, int unit)
{
    int i, j;
    
    #define SCE_DELETE_TEX(a, b)\
    {\
        if(!mat->maps[a][b].user)\
            SCE_Texture_Delete(mat->maps[a][b].tex);\
        mat->maps[a][b].tex = NULL;\
        mat->maps[a][b].actived = SCE_FALSE;\
        mat->maps[a][b].user = SCE_FALSE;\
    }
    
    if(mat != NULL)
    {
        if(type == SCE_ALL_MAPS)
        {
            for(i=0; i<SCE_NUM_MAPS; i++)
            {
                if(unit == SCE_ALL_MAPS)
                {
                    for(j=0; j<SCE_CGetInteger(GL_MAX_TEXTURE_UNITS); j++)
                        SCE_DELETE_TEX(i, j)
                }
                else
                    SCE_DELETE_TEX(i, unit)
            }
        }
        else if(unit == SCE_ALL_MAPS)
        {
            for(j=0; j<SCE_CGetInteger(GL_MAX_TEXTURE_UNITS); j++)
                SCE_DELETE_TEX(type, j)
        }
        else
            SCE_DELETE_TEX(type, unit)
    }
    
    #undef SCE_DELETE_TEX
}

/* ajoute le 14/06/2007 */
/* revise le 18/06/2007 */
void SCE_Material_DeleteShader(SCE_SMaterial *mat)
{
    if(mat != NULL)
    {
        if(!mat->shader_user)
            SCE_Shader_Delete(mat->shader);
        
        mat->shader = NULL;
        mat->shader_actived = SCE_FALSE;
        mat->shader_user = SCE_FALSE;
    }
}


/* ajoute le 14/06/2007 */
/* revise le 19/06/2007 */
/** fonction a tester */
void SCE_Material_ActiveTexture(SCE_SMaterial *mat, int type, int id, int state)
{
    int i, j;
    
    #define SCE_ACTIVE_TEX(a, b)\
    {\
        /* uniquement si ce n'est pas une texture inexistante */\
        if(mat->maps[a][b].tex != NULL)\
            mat->maps[a][b].actived = state;\
    }
    
    if(mat != NULL)
    {
        if(type <= SCE_ALL_MAPS)
        {
            for(i=0; i<SCE_NUM_MAPS; i++)
            {
                if(id == SCE_ALL_MAPS)
                {
                    for(j=0; j<SCE_CGetInteger(GL_MAX_TEXTURE_UNITS); j++)
                        SCE_ACTIVE_TEX(i, j)
                }
                else
                    SCE_ACTIVE_TEX(i, id)
            }
        }
        else if(id == SCE_ALL_MAPS)
        {
            for(j=0; j<SCE_CGetInteger(GL_MAX_TEXTURE_UNITS); j++)
                SCE_ACTIVE_TEX(type, j)
        }
        else
            SCE_ACTIVE_TEX(type, id)
    }
}

/* ajoute le 14/06/2007 */
void SCE_Material_ActiveShader(SCE_SMaterial *mat, int actived)
{
    if(mat != NULL)
    {
        if(mat->shader != NULL)
            mat->shader_actived = actived;
    }
}


/* revise le 18/06/2007 */
void SCE_Material_Use(SCE_SMaterial *mat)
{
    int i, j;
    
    if(mat != NULL)
    {
        /* desactivation de toutes les unites de texture */
        for(i=0; i<SCE_CGetInteger(GL_MAX_TEXTURE_UNITS); i++)
            SCE_Texture_Use(NULL, i);
        
        for(i=0; i<SCE_NUM_MAPS; i++)
        {
            for(j=0; j<SCE_CGetInteger(GL_MAX_TEXTURE_UNITS); j++)
            {
                if(mat->maps[i][j].actived)
                    SCE_Texture_Use(mat->maps[i][j].tex, j);
            }
        }
        
        if(mat->shader_actived)
            SCE_Shader_Use(mat->shader);
        else
            SCE_Shader_Use(NULL);
        
        SCE_CUseMaterial(mat->cmat);
    }
    else
    {
        SCE_CUseMaterial(NULL);
        SCE_Shader_Use(NULL);
        
        for(i=0; i<SCE_CGetInteger(GL_MAX_TEXTURE_UNITS); i++)
            SCE_Texture_Use(NULL, i);
    }
}
